package com.chelvc.framework.kafka.fallback;

import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

import com.chelvc.framework.common.util.ThreadUtils;
import com.chelvc.framework.kafka.producer.KafkaMessageSender;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.springframework.kafka.support.SendResult;
import org.springframework.util.concurrent.ListenableFuture;
import org.springframework.util.concurrent.SettableListenableFuture;

/**
 * Kafka降级内存发送处理器实现
 *
 * @param <K> 顺序标识类型
 * @param <V> 消息载体类型
 * @author Woody
 * @date 2024/7/18
 */
@Slf4j
public class KafkaMemorySender<K, V> implements KafkaMessageSender<K, V> {
    private final Thread thread;
    private final BlockingQueue<ProducerRecord<K, V>> queue;
    private final KafkaMessageSender<K, V> delegate;

    public KafkaMemorySender(@NonNull KafkaMessageSender<K, V> delegate) {
        this.delegate = delegate;
        this.queue = new ArrayBlockingQueue<>(100000);
        this.thread = ThreadUtils.consume(this.queue, record -> {
            try {
                this.delegate.send(record);
                return true;
            } catch (Exception e) {
                log.error("Kafka message send failed: {}", record.topic(), e);
                ThreadUtils.sleep(5000);
            }
            return false;
        });
    }

    /**
     * 消息发送降级处理
     *
     * @param record 消息记录
     */
    private void fallback(ProducerRecord<K, V> record) {
        if (this.queue.offer(record)) {
            if (log.isDebugEnabled()) {
                log.debug("Kafka fallback memory message: {}, {}", record.topic(), record.value());
            }
            return;
        }
        throw new RuntimeException(String.format(
                "Kafka fallback memory queue is full. Message discarded: %s, %s", record.topic(), record.value()
        ));
    }

    @Override
    public ListenableFuture<SendResult<K, V>> send(ProducerRecord<K, V> record) {
        try {
            return this.delegate.send(record);
        } catch (Exception e) {
            log.error("Kafka message send failed: {}", record.topic(), e);
        }

        // 降级处理
        this.fallback(record);
        SettableListenableFuture<SendResult<K, V>> future = new SettableListenableFuture<>();
        future.set(new FallbackSendResult<K, V>(record));
        return future;
    }

    @Override
    public void destroy() throws Exception {
        this.thread.interrupt();
        ThreadUtils.join(this.thread);
    }
}
