package com.chelvc.framework.rocketmq.fallback;

import java.io.IOException;
import java.util.concurrent.CompletableFuture;

import com.chelvc.framework.base.context.Session;
import com.chelvc.framework.base.context.SessionContextHolder;
import com.chelvc.framework.common.util.JacksonUtils;
import com.chelvc.framework.kafka.context.KafkaContextHolder;
import com.chelvc.framework.rocketmq.context.RocketMQContextHolder;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.rocketmq.client.apis.ClientException;
import org.apache.rocketmq.client.apis.message.Message;
import org.apache.rocketmq.client.apis.producer.Producer;
import org.apache.rocketmq.client.apis.producer.SendReceipt;
import org.apache.rocketmq.client.apis.producer.Transaction;

/**
 * RocketMQ降级Kafka消息生产者实现
 *
 * @author Woody
 * @date 2024/7/13
 */
@Slf4j
public class RocketMQKafkaProducer implements Producer {
    private final Producer delegate;

    public RocketMQKafkaProducer(@NonNull Producer delegate) {
        this.delegate = delegate;
    }

    /**
     * 消息发送降级处理
     *
     * @param message 消息实例
     */
    private void fallback(Message message) {
        if (log.isDebugEnabled()) {
            String topic = RocketMQContextHolder.topic(message), body = RocketMQContextHolder.body(message);
            log.debug("RocketMQ fallback kafka message: {}, {}", topic, body);
        }
        String topic = RocketMQContextHolder.topic(message, '-');
        Object payload = JacksonUtils.deserialize(RocketMQContextHolder.body(message));
        String ordering = message.getMessageGroup().orElse(null);
        Session session = JacksonUtils.deserialize(message.getProperties().get(Session.NAMING), Session.class);
        SessionContextHolder.setSession(session);
        try {
            if (ordering == null) {
                KafkaContextHolder.send(topic, payload);
            } else {
                KafkaContextHolder.send(topic, payload, ordering);
            }
        } finally {
            SessionContextHolder.clearSessionContext();
        }
    }

    @Override
    public SendReceipt send(Message message) throws ClientException {
        try {
            return this.delegate.send(message);
        } catch (Exception e) {
            log.error("RocketMQ message send failed: {}", RocketMQContextHolder.topic(message), e);
        }

        // 降级处理
        this.fallback(message);
        return FallbackSendReceipt.INSTANCE;
    }

    @Override
    public SendReceipt send(Message message, Transaction transaction) throws ClientException {
        return this.delegate.send(message, transaction);
    }

    @Override
    public CompletableFuture<SendReceipt> sendAsync(Message message) {
        return this.delegate.sendAsync(message).exceptionally(e -> {
            log.error("RocketMQ message async send failed: {}", RocketMQContextHolder.topic(message), e);

            // 降级处理
            this.fallback(message);
            return FallbackSendReceipt.INSTANCE;
        });
    }

    @Override
    public Transaction beginTransaction() throws ClientException {
        return this.delegate.beginTransaction();
    }

    @Override
    public void close() throws IOException {
        this.delegate.close();
    }
}
