package com.chelvc.framework.rocketmq.fallback;

import java.io.IOException;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.CompletableFuture;

import com.chelvc.framework.common.util.ThreadUtils;
import com.chelvc.framework.rocketmq.context.RocketMQContextHolder;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.rocketmq.client.apis.ClientException;
import org.apache.rocketmq.client.apis.message.Message;
import org.apache.rocketmq.client.apis.producer.Producer;
import org.apache.rocketmq.client.apis.producer.SendReceipt;
import org.apache.rocketmq.client.apis.producer.Transaction;

/**
 * RocketMQ降级内存消息生产者实现
 *
 * @author Woody
 * @date 2024/7/14
 */
@Slf4j
public class RocketMQMemoryProducer implements Producer {
    private final Thread thread;
    private final Producer delegate;
    private final BlockingQueue<Message> queue;

    public RocketMQMemoryProducer(@NonNull Producer delegate) {
        this.delegate = delegate;
        this.queue = new ArrayBlockingQueue<>(100000);
        this.thread = ThreadUtils.consume(this.queue, message -> {
            try {
                this.delegate.send(message);
                return true;
            } catch (Exception e) {
                log.error("RocketMQ message send failed: {}", RocketMQContextHolder.topic(message), e);
                ThreadUtils.sleep(5000);
            }
            return false;
        });
    }

    /**
     * 消息发送降级处理
     *
     * @param message 消息实例
     */
    private void fallback(Message message) {
        if (this.queue.offer(message)) {
            if (log.isDebugEnabled()) {
                String topic = RocketMQContextHolder.topic(message), body = RocketMQContextHolder.body(message);
                log.debug("RocketMQ fallback memory message: {}, {}", topic, body);
            }
            return;
        }
        String topic = RocketMQContextHolder.topic(message), body = RocketMQContextHolder.body(message);
        throw new RuntimeException(String.format(
                "RocketMQ fallback memory queue is full. Message discarded: %s, %s", topic, body
        ));
    }

    @Override
    public SendReceipt send(Message message) throws ClientException {
        try {
            return this.delegate.send(message);
        } catch (Exception e) {
            log.error("RocketMQ message send failed: {}", RocketMQContextHolder.topic(message), e);
        }

        // 降级处理
        this.fallback(message);
        return FallbackSendReceipt.INSTANCE;
    }

    @Override
    public SendReceipt send(Message message, Transaction transaction) throws ClientException {
        FallbackTransaction proxy = (FallbackTransaction) transaction;
        if (!proxy.isDegraded()) {
            try {
                return this.delegate.send(message, proxy.getTransaction());
            } catch (Exception e) {
                log.error("RocketMQ transactional message send failed: {}", RocketMQContextHolder.topic(message), e);
            }
        }

        // 降级处理
        this.fallback(message);
        proxy.delegate(message);
        return FallbackSendReceipt.INSTANCE;
    }

    @Override
    public CompletableFuture<SendReceipt> sendAsync(Message message) {
        return this.delegate.sendAsync(message).exceptionally(e -> {
            log.error("RocketMQ message async send failed: {}", RocketMQContextHolder.topic(message), e);

            // 降级处理
            this.fallback(message);
            return FallbackSendReceipt.INSTANCE;
        });
    }

    @Override
    public Transaction beginTransaction() throws ClientException {
        Transaction transaction = null;
        try {
            transaction = this.delegate.beginTransaction();
        } catch (Exception e) {
            log.error("RocketMQ transaction open failed", e);
        }
        return new FallbackTransaction(transaction, this::fallback);
    }

    @Override
    public void close() throws IOException {
        try {
            this.thread.interrupt();
            ThreadUtils.join(this.thread);
        } finally {
            this.delegate.close();
        }
    }
}
